@@ -3,17 +3,13 @@ package Algorithm::RabinKarp;
 use warnings;
 use strict;
 
-use Algorithm::RabinKarp::Util qw(stream_fh stream_string);
+use constant BASE => 101;
 
-use UNIVERSAL;
-
-use constant BASE => 2;
-
-our $VERSION = "0.41_1";
+our $VERSION = "0.33";
 
 =head1 NAME
 
-Algorithm::RabinKarp - Rabin-Karp streaming hash
+Algorithm::RabinKarp - rabin-karp streaming hash
 
 =head1 SYNOPSIS
 
@@ -30,7 +26,7 @@ or
     return $num, $position;
   });
   
-  my ($hash, $start_position, $end_position) = $kgram->next;
+  my ($first, $start_position, $end_position) = $kgram->next;
   
   my @values = $kgram->values;
   
@@ -69,20 +65,6 @@ your data to remove all unnecessary information. For example, in a large
 english document, you should probably remove all white space, as well
 as removing all capitalization.
 
-=head1 INTENT
-
-By preprocessing your document with the Rabin Karp hashing algorithm,
-it makes it possible to create a "fingerprint" of your document (or documents), 
-and then perform multiple searches for fragments contained within your document
-database.
-
-Schleimer, Wilkerson, and Aiken suggest preproccessing to remove
-unnecessary information (like whitespace), as well as known redundent information
-(like, say, copyright notices or other boilerplate that is 'acceptable'.)
-
-They also suggest a post processing pass to reduce data volume, using a technique
-called winnowing (see the link at the end of this documentation.)
-
 =head1 METHODS
 
 =over
@@ -99,45 +81,49 @@ filtering characters out because they're redundant.)
 sub new {
   my $class = shift;
   my $k = shift;
-  my $stream = $class->make_stream(shift); 
-  my $rm_k = BASE;
+  my $source;
+  if (!ref $_[0]) {
+    open $source, '<', \$_[0] 
+      or die "Couldn't create a file handle on scalar source: $!";
+  } else {
+    $source = $_[0];
+  }
+  my $stream;
+  if ($source->isa("IO::Handle") || (ref $source) =~  /^GLOB/) {
+    require IO::Handle;
+    # The simplest way of getting character position right now.
+    my $counter = 0;
+    $stream = sub {
+      return if $source->eof;
+      (ord($source->getc), $counter++);
+    };
+  } elsif (ref $source eq 'CODE') {
+    $stream = $source;
+  }
+  
+  die __PACKAGE__." requires either a scalar, file handle, or coderef"
+    unless $source;
+    
+  if (BASE ** $k <= 0) {
+    require bignum;
+  }
   
   bless { 
-    k => $k % 32,
+    k => $k,
+    rm_k => BASE ** $k, #used to remove the first value in the value buffer.
     vals => [],
     stream => $stream,
   }, ref $class || $class;
 }
 
-sub make_stream {
-  my $class = shift;
-  my $source = shift;	
-  
-  return $source if ref $source eq 'CODE';
-  
-  my $stream; 
-  if (defined $source && !ref $source) {
-    $stream = stream_string($source);
-  } elsif (UNIVERSAL::isa($source, "IO::Handle") 
-           || UNIVERSAL::isa($source,"GLOB")) {
-    require IO::Handle;
-    # The simplest way of getting character position right now.
-    $stream = stream_fh($source);
-  } else {
-    die __PACKAGE__." requires its source stream be one of the ".
-        "following types: scalar, file handle, coderef, or IO::Handle";
-  }
-  return $stream;
-}
-  
-
 =item next()
 
-Returns an array containing (kgram hash value, start position , end position, start, end) 
-for every call that can have a hash generated, or () when we have reached the end
-of the stream.  
+Returns an array of three values for each call. The first element is
+the k-gram hash value.  The second and third elements are the start and
+end positions, inclusive, as provided by the generator stream.
 
-C<next()> pulls the first $k from the stream on the first call. Each successive
+C<next()> requires $k iterations to warm up on first call (don't worry, you
+don't need to remember to do that, it's handled internally.)  Each successive
 call to C<next()> has a complexity of O(1).
 
 =cut
@@ -146,22 +132,25 @@ sub next {
 
   # assume, for now, that each value is an integer, or can
   # auto cast to char
-  my $values = $self->{vals}; #assume that @values always contains k values
-  my $prev = shift @$values || [0, undef];
-  my $hash = $self->{hash} || 0;
-  while (@$values < $self->{k}) {
+  my @values = @{$self->{vals}}; #assume that @values always contains k values
+  my $prev = shift @values || [0, undef];
+  my $hash = $self->{hash};
+  do {
     my $nextval = [$self->{stream}->()];
     return unless @$nextval;
-    push @$values, $nextval;
-    $hash <<= 1;
-    $hash -= $prev->[0] << $self->{k};
-    $hash += $nextval->[0];
     
-  }
-
+    push @values, $nextval;
+    # If someone wants to submit a modulus version of this
+    # I would be most grateful.
+    $hash -= $prev->[0] * $self->{rm_k};
+    $hash += $nextval->[0];
+    $hash *= BASE;
+  } while (@values < $self->{k});
+  #warn join( '', map { chr($_ ) } @values). ' '.$hash;
   $self->{hash} = $hash;
+  $self->{vals} = \@values;
   
-  return $hash, $values->[0][1], $values->[-1][1], @{ $values }[0, -1];
+  return $hash, $values[0]->[1], $values[-1]->[1];
 }
 
 =item values
@@ -173,8 +162,8 @@ format as yielded by L<next|/METHODS>.)
 After calling C<values()> the stream will be completely exhausted, causing 
 subsequent calls to C<values> and C<next()> to return C<undef>.
 
-NOTE: You should use C<next> if your source stream is infinite, as values
-will greedily attempt to consume all values.
+NOTE: You should use C<next> if the stream you are generating hash codes for
+is infinite. Failure to do so will yield unexpected results.
 
 =cut
 
@@ -194,18 +183,17 @@ sub values {
 
 =head1 BUGS
 
-The current multipliers and modulus lead to very poor hash
-distributions.  I'll investigate methods of improving this
-in future versions.
+No explicit guards against overflow have been taken, nor any attempts to
+use clever bitwise operators.  I recommend either using small values of
+C<k>, or including a 'use bignum' line before before iterating with larger
+C<k>s.  In a future version of this module, I will shift to using modulo
+arithmetic, which will not have the same sort of overflow problems.
 
 =head1 SEE ALSO
 
   "Winnowing: Local Algorithms for Document Fingerprinting"
   L<http://theory.stanford.edu/~aiken/publications/papers/sigmod03.pdf>
 
-  Wikipedia: Rabin-Karp string search algorithm
-  L<http://en.wikipedia.org/wiki/Rabin-Karp>
-
 =head1 AUTHOR
 
   Norman Nunley E<lt>nnunley@gmail.comE<gt>
